""" Database adapters """
from typing import List, Dict, Tuple
from abc import ABC, abstractmethod
class AbstractAdapter(ABC):
    """
    The main and only source for building the schema.

    All the methods in this interface are mandatory. If you do not want to add a specific element
    to the schema to be built, implement the given function returning an empty list([]).
    Note, however, that not putting elements that are related to others, namely tables and
    columns, it may cause unexpected behavior

    The methods can be classified in two types: entity methods and relation method.

    Entity Methods are the ones that are used to generate an entity(table, column, etc.),
    and they need to return a list of dicts.
    Many of these methods also need to inform its relation to another entities, but its main purpose
    is to inform what are the entities in the database and what properties
    should this given object have, and the values it returns.

    Entity methods have a dict to represent an entity because it allows you to pass
    any kind of parameter of an object you want,
    at the same it that it does not enforce any of them besides its name
    and its relations to another entities. So, for example,
    if you want to make a validation that utilizes its data type,
    add a key which value is the column data type ('data_type', for example),
    and you will be able to access
    this information through the object as a method (column.data_type, for example).

    Relation methods are, in the other hand, used to inform certain relationships between entities,
    specifically primary and foreign keys.
    This information is presented separated from the column method
    in order to not insert a disproportional amount of logic inside one method(columns).
    They need to return a list of tuples, or at least in an ordered object(like a list)
    the information in what needs to appear
    and in which order is explained in more detail in each method docstring.
    """
    @abstractmethod
    def tables(self) -> List[Dict[str, str]]:
        """Return a list containing a dict for every table in the schema

         mandatory keys:
          - name: The Table name
         forbidden keys:
          - columns
          - triggers
        """
        raise NotImplementedError

    @abstractmethod
    def columns(self) -> List[Dict[str, str]]:
        """Return a dict for every column in the schema

         mandatory keys:
          - name: The Column name
          - table: The name of the Table that owns the column
         forbiddent keys:
          - constraints
          - indexes
          - primary_keys
          - references
        """
        raise NotImplementedError

    @abstractmethod
    def triggers(self) -> List[Dict[str, str]]:
        """Return a list of dicts, each representing one of every triggers in the schema

        mandatory keys:
         - name: the Trigger name
         - table: The Name of the Table that owns it
        """
        raise NotImplementedError

    @abstractmethod
    def primary_keys(self) -> List[Tuple[str, str]]:
        """Return a tuple for every table that contains one primary key.
        If the said table has no primary key, do not put the table at all.

        The tuple must have the table name as its first element, and the name
        of the column that is its primary key as it's second. Currently does not
        support primary keys with multiple columns.
        """
        raise NotImplementedError

    @abstractmethod
    def references(self) -> List[Tuple[str, str, str]]:
        """Return a tuple for every column that is a foreign key for a table, ordered by:

        (the name of the column table, the column name, the name of the referenced table)
        If a column is not a reference for any table, it should not appear in this list at all.
        """
        raise NotImplementedError

    @abstractmethod
    def indexes(self) -> List[Dict[str, str]]:
        """ Return a dict for every index int the schema

        mandatory keys:
         - name: the Index name
         - column: the column bound with the index
        currently, it does not support a multi-column index """
        raise NotImplementedError

    @abstractmethod
    def constraints(self) -> List[Dict[str, str]]:
        """Return a dict for every constraint present in the schema

        mandatory keys:
         - name: the Constraint name
         - column: the column bound with constraint
        """
        raise NotImplementedError

    @abstractmethod
    def functions(self) -> List[Dict[str, str]]:
        """Return a dict for every function present in the schema

        mandatory keys:
         - name: the function name
        """
        raise NotImplementedError

    @abstractmethod
    def procedures(self) -> List[Dict[str, str]]:
        """Return a dict for every procedure present in the schema

        mandatory keys:
         - name: the procedure name
        """
        raise NotImplementedError

    @abstractmethod
    def sequences(self) -> List[Dict[str, str]]:
        """Return a dict for every sequence present in the schema

        mandatory keys:
         - name: the sequence name
        """
        raise NotImplementedError



""" Generate Reports in different output format """
from typing import List, Dict, Callable, Any
from .csv import export_csv
from .cli import export_cli

def formatted_output(report: dict, xformat: str) -> List[str]:
    """Generates a report based in the format presented in 'xformat'"""
    formats: Dict[str, Callable[[Any], List[str]]] = {
        'CLI': export_cli,
        'CSV': export_csv
    }
    return formats[xformat](report)



"""Test Utils"""

from typing import Dict, List, Callable
from .parse_configuration.schema_validations import module_validations, to_configuration
from .validate import validate_entity
from .schema import Entity

class Validator:
    """Allows to validate a single entity against a validation module"""
    def __init__(self, validations: Dict[str, List[Callable]]):
        self.validations = to_configuration(module_validations(validations))

    def valid(self, entity: Entity):
        """Checks if entity is valid against the validations"""
        return len(self.validate(entity)) == 0

    def validate(self, entity: Entity):
        """Returns all invalidations of an entity. Returns an empty list if there is none"""
        return [v for v in validate_entity(self.validations, entity) if v is not None]



""" Generate the Report Output to the schema validation

Report Example:

REPORT
=================================================

Tables:
=================================================
 + table_one
   + message one
   + message two
---------------------------------------

Columns:
=================================================
 + table_one.column_one
   + message one
---------------------------------------
 + table_one.column_two
   + message one
---------------------------------------
"""
from typing import List, Dict
from ..util import group_by

def _entity_report(entity_name: str, messages: List[str]) -> List[str]:
    """ Serializes the messages of a given entity to the Report

    It generates a list that translates as:

     + entity
       + message one
       + message two"""
    return [f' + {entity_name}'] + [f'   + {message}' for message in messages]

def _entities_report(entity_group: str, entities: Dict[str, List[str]]) -> List[str]:
    """ Serializes the messages to a group of entities to the Report """
    output = [f'{entity_group}:', '=' * 50]
    for entity, messages in entities.items():
        output += _entity_report(entity, messages)
        output.append('-' * 40)
    return output

# entities_report: { 'entity_group': { 'entity': ['message'] } }
def export_cli(report_hash: Dict[str, Dict[str, List[str]]]) -> List[str]:
    """ Receives an Dict containing the Report info, divided in:
        key -> The Entity Group (Ex.: Table, Columns)
        value -> A Dict containing the Entities and its messages
            key -> The Entity unique name (Ex.: table_one, table_one.column_one)
            value -> A list of Strings containing the Validations Mistakes in the Report"""
    output = ['REPORT', '=' * 50]

    ygroups = group_by(report_hash, lambda x: x.group)
    groups = {key:group_by(value, lambda x: x.report_name, lambda x: x.message) for key, value in ygroups.items()}

    for entity_group, entities in groups.items():
        output += _entities_report(entity_group, entities)
    return output



"""Confinguration Builder for Adapter Options"""
from abc import ABC, abstractmethod
from typing import List, Dict
import importlib
import pkg_resources

class AdapterBuilder(ABC):
    """Abstract Class for retrieving the adapter used to fetch DB schema data"""
    params: List[str]
    named_params: Dict[str, str]

    def __init__(self, params = None, named_params = None):
        self._invalidation = None
        self.params = params or []
        self.named_params = named_params or {}

    @classmethod
    @abstractmethod
    def elligible(cls, config) -> bool:
        """Checks whether the config matches with the type"""

    @abstractmethod
    def asdict(self) -> dict:
        """Convert Params to a Dict"""

    def merge(self, other) -> 'AdapterBuilder':
        """Merge two Adapter Options"""
        keys = {*self.asdict(), *other.asdict()}
        ret = {}
        for key in keys:
            if key == 'named_params':
                continue
            ret[key] = other.asdict().get(key) or self.asdict().get(key)
        ret['named_params'] = {**self.named_params, **other.named_params}
        return from_json(ret)   

    @abstractmethod
    def build(self):
        """Build Configuration Adapter Param"""

    @abstractmethod
    def is_valid(self) -> bool:
        """Checks validity"""

    def __eq__(self, other):
        return self.asdict() == other.asdict()

    def error(self) -> str:
        """Invalid builder message"""
        return self._invalidation

class UnresolvedAdapterBuilder(AdapterBuilder):
    """Similar to a Null Object, it is instantiaded in a 'from_json' call whenever
    it cannot resolve to an appended or pluggable type."""
    def __init__(self, params = None, named_params = None, **_unresolved):
        super().__init__(params, named_params)

    @classmethod
    def elligible(cls, _config):
        return True

    def is_valid(self):
        self._invalidation = 'Builder could not resolve Adapter Type'
        return False

    def asdict(self):
        return {'params': self.params, 'named_params': self.named_params}

    def build(self):
        raise RuntimeError('Cannot Build an Unresolved Adapter Builder')

class AppendedAdapterBuilder(AdapterBuilder):
    """Adapter Builder for modules used in Python discovery"""
    def __init__(self, module, klass, params = None, named_params = None):
        super().__init__(params, named_params)
        self.module = module
        self.klass = klass

    @classmethod
    def elligible(cls, config):
        return 'module' in config and 'klass' in config

    def is_valid(self):
        if self.module is None:
            self._invalidation = 'Adapter Module not provided'
            return False
        if self.klass is None:
            self._invalidation = 'Adapter Class not provided'
            return False
        return True

    def asdict(self):
        return {
            'module': self.module,
            'klass': self.klass,
            'params': self.params,
            'named_params': self.named_params
        }

    def build(self):
        adapter_module = importlib.import_module(self.module)
        return getattr(adapter_module, self.klass)(*self.params, **self.named_params)

class PluggableAdapterBuilder(AdapterBuilder):
    """Adapter Builder for modules that are plugins"""
    def __init__(self, plugin, params = None, named_params = None):
        super().__init__(params, named_params)
        self.plugin = plugin

    @classmethod
    def elligible(cls, config):
        return 'plugin' in config

    def is_valid(self):
        if self.plugin is None:
            return False
        return True

    def asdict(self):
        return {
            'plugin': self.plugin,
            'params': self.params,
            'named_params': self.named_params
        }

    def build(self):
        iter_adapter_plugins = pkg_resources.iter_entry_points('sql_judge.adapter')
        try:
            plugin_module = next(p for p in iter_adapter_plugins if p.name == self.plugin).load()
            return getattr(plugin_module, 'Adapter')(*self.params, **self.named_params)
        except StopIteration as stopit:
            raise RuntimeError(f"Could not find plugin with '{self.plugin}' ID") from stopit

def from_json(options):
    """AdapterBuilder factory made out of a dict.
    The Type of Adapter depends on the keys the config holds"""
    if 'class' in options:
        options['klass'] = options['class']
        del options['class']
    for builder in (AppendedAdapterBuilder,PluggableAdapterBuilder):
        if builder.elligible(options):
            return builder(**options)
    return UnresolvedAdapterBuilder(**options)

def default_adapter():
    """Default Adapter Configuration. Alias for an adapter that has no properties"""
    return from_json({})



"""Database Schema"""
from typing import List, Optional
from collections import namedtuple
from .util import find, cached_property

class Schema:
    """Database Schema"""
    def __init__(self, adapter):
        self._adapter = adapter

        # Caching references used more than once to avoid making multiple calls to the adapter
        self._references = self._adapter.references()
        self._primary_keys = self._adapter.primary_keys()

    def __references(self, params):
        ref_tuple = namedtuple('Reference', ['table', 'column', 'references'])
        references = (ref_tuple(*el) for el in self._references)
        find_reference = lambda el: el.table == params['table_name'] and el.column == params['name']
        ref = find(references, find_reference)
        return ref.references if ref is not None else None

    def __is_primary_key(self, params):
        pkey = find(
            self._primary_keys,
            lambda el: el == (params['table_name'], params['name'])
        )
        return pkey is not None

    def __entities(self, factory, group):
        return [
            factory(schema=self, **params) for params in getattr(self._adapter, group)()
        ]

    @cached_property
    def tables(self) -> List['Table']:
        """ Database Tables"""
        return self.__entities(Table, 'tables')

    @cached_property
    def columns(self) -> List['Column']:
        """ Database Columns """
        return [
            Column(
                schema=self, primary_key = self.__is_primary_key(params),
                references = self.__references(params), **params)
            for params in self._adapter.columns()
        ]

    @cached_property
    def sequences(self) -> List['Entity']:
        """ Database Sequences """
        return self.__entities(Sequence, 'sequences')

    @cached_property
    def functions(self) -> List['Entity']:
        """ Database Functions """
        return self.__entities(Function, 'functions')

    @cached_property
    def procedures(self) -> List['Entity']:
        """ Database Procedures """
        return self.__entities(Procedure, 'procedures')

    @cached_property
    def triggers(self) -> List['TableEntity']:
        """ Database Table Triggers """
        return self.__entities(Trigger, 'triggers')

    @cached_property
    def constraints(self) -> List['ColumnEntity']:
        """ Database Column Constraints """
        return self.__entities(Constraint, 'constraints')

    @cached_property
    def indexes(self) -> List['ColumnEntity']:
        """ Database Column Indexes """
        return self.__entities(Index, 'indexes')

    def entities(self):
        """ Schema Entities """
        return [
            *self.tables,
            *self.columns,
            *self.triggers,
            *self.indexes,
            *self.constraints,
            *self.sequences,
            *self.functions,
            *self.procedures
        ]


class Entity:
    """ Generic Database Schema Entity """
    def __init__(self, group: str, schema: Schema, name: str, **custom_params):
        self._schema: Schema = schema
        self.__name: str = name
        self._custom_params: dict = custom_params
        self.__name__ = group.capitalize()

    @property
    def name(self) -> str:
        """ Entity Name """
        return self.__name

    @property
    def schema(self) -> Schema:
        """ Entity Schema """
        return self._schema

    def __getattribute__(self, item):
        if item in super().__getattribute__('_custom_params'):
            return self._custom_params[item]
        return super().__getattribute__(item)

    def __str__(self):
        return f'<{self.__name__} "{self.name}">'

    def __repr__(self):
        return self.__str__()


def Function(*args, **kwargs): # pylint: disable=invalid-name
    """Function Entity"""
    return Entity(group='function', *args, **kwargs)

def Procedure(*args, **kwargs): # pylint: disable=invalid-name
    """Procedure Entity"""
    return Entity(group='procedure', *args, **kwargs)

def Sequence(*args, **kwargs): # pylint: disable=invalid-name
    """Sequence Entity"""
    return Entity(group='sequence', *args, **kwargs)

class Table(Entity):
    """Database Table"""
    def __init__(self, name, schema, **custom_params):
        super().__init__(group='table', name=name, schema=schema, **custom_params)

    @cached_property
    def columns(self) -> List['Column']:
        """ Table Columns """
        return [
            column for column in self._schema.columns if column.table.name == self.name
        ]

    @cached_property
    def primary_key(self) -> Optional['Column']:
        """ Table Primary Key Column """
        return find(self.columns, lambda col: col.primary_key)

    @cached_property
    def triggers(self) -> List['TableEntity']:
        """ Table Triggers """
        return [
            trigger for trigger in self._schema.triggers if trigger.table.name == self.name
        ]

class TableEntity(Entity):
    """ Entity That belongs to a Table """
    def __init__(self, group, name, table_name, schema, **custom_params):
        super().__init__(group=group, name=name, schema=schema, **custom_params)
        self._table = find(self._schema.tables, lambda el: el.name == table_name)

    @property
    def table(self) -> Table:
        """ Table assigned to Entity """
        return self._table

def Trigger(**params): # pylint: disable=invalid-name
    """ Trigger Entity """
    return TableEntity(group='trigger', **params)

class Column(TableEntity):
    """ Column Entity """
    def __init__(self, name, table_name, schema, primary_key=False, references=None, **custom_params):
        super().__init__(group='column', name=name, table_name=table_name, schema=schema, **custom_params)
        self.__primary_key: bool = primary_key
        self._references: str = references

    @property
    def primary_key(self) -> bool:
        """ Is Primary Key? """
        return self.__primary_key

    @property
    def references(self) -> Optional[Table]:
        """ The table it references, if any """
        if not self._references:
            return None
        return find(self._schema.tables, lambda el: el.name == self._references)

    @property
    def index(self) -> Optional['ColumnEntity']:
        """ Column Indexes """
        return find(
            self._schema.indexes,
            lambda index: index.table == self.table and index.column.name == self.name
        )

    @property
    def constraints(self) -> List['ColumnEntity']:
        """ Column Constraint """
        return [
            cons for cons in self._schema.constraints
            if cons.table == self.table and cons.column.name == self.name
        ]

class ColumnEntity(TableEntity):
    """ Entity that belongs to a Column """
    def __init__(self, group, name, table_name, column_name, schema, **custom_params):
        super().__init__(group=group, name=name, table_name=table_name, schema=schema, **custom_params)
        self._column = find(
            self._schema.columns,
            lambda el: el.table == self.table and el.name == column_name
        )

    @property
    def column(self) -> Column:
        """ Column Assigned to Entity """
        return self._column

def Index(*args, **kwargs): # pylint: disable=invalid-name
    """ Index Entity """
    return ColumnEntity(group='index', *args, **kwargs)

def Constraint(*args, **kwargs): # pylint: disable=invalid-name
    """ Constraint Entity """
    return ColumnEntity(group='constraint', *args, **kwargs)



""" Sets ConfigurationBuilder, a class responsible to merge configuration
options from various sources, merge them, and build a valid Configuration
Object to use in the rest of the proccess """
import importlib
import json
from dataclasses import dataclass, field
from typing import List, Optional
from .. import Configuration
from . import schema_validations, adapter_builder


@dataclass
class ConfigurationBuilder:
    """Group, Validate and Build the Configurations parameters"""
    adapter: adapter_builder.AdapterBuilder = field(default_factory=adapter_builder.default_adapter)
    validations_module: Optional[str] = None
    ignore_tables: List[str] = field(default_factory=list)
    export_format: Optional[str] = None
    export_output: Optional[str] = None
    _invalidation: Optional[str] = None

    @classmethod
    def from_json(cls, json_str):
        """Generates an instance based from a JSON string"""
        try:
            options = json.loads(json_str)
        except json.JSONDecodeError as jserr:
            raise RuntimeError('Error while parsing configuration') from jserr
        adapter_options = options.get('adapter', {})
        validation_options = options.get('validations', {})
        export_options = options.get('export', {})
        return cls(
            adapter = adapter_builder.from_json(adapter_options),
            ignore_tables=options.get('ignore_tables', []),
            validations_module=validation_options.get('module'),
            export_format=export_options.get('format')
        )

    def merge(self, builder: 'ConfigurationBuilder'):
        """ Combine two builders """
        return ConfigurationBuilder(
            adapter=self.adapter.merge(builder.adapter),
            validations_module=builder.validations_module or self.validations_module,
            ignore_tables=self.ignore_tables + builder.ignore_tables,
            export_format=builder.export_format or self.export_format
        )

    def is_valid(self):
        """Checks validity of Configuration Builder"""
        if not self.adapter.is_valid():
            self._invalidation = self.adapter.error()
            return False
        if self.validations_module is None:
            self._invalidation = 'Validations Module not provided'
            return False
        if self.export_format is None:
            self._invalidation = 'Export Format not provided'
            return False
        return True


    def _validations(self):
        validations_module = importlib.import_module(self.validations_module)
        validations = schema_validations.module_validations(validations_module)
        return schema_validations.to_configuration(validations)

    def build(self) -> Configuration:
        """Use the internal parameters to generate a Configuration Instance"""
        if not self.is_valid():
            raise RuntimeError(self._invalidation)
        return Configuration(
            connection=self.adapter.build(),
            validations=self._validations(),
            export=self.export_format, #type: ignore
            ignore_tables=self.ignore_tables
        )



""" Generate the Report Output in CSV format

Report Example:
Table, table_one, message one
Table, table_one, message two
Column, table_one.column_one, message one
Column, table_one.column_two, message one
"""
from typing import List
from ..validate import Fail
def export_csv(report_hash: List[Fail]) -> List[str]:
    report = []
    for group, report_name, message in report_hash:
        report.append(f'{group.capitalize()}, {report_name}, {message}')
    return report



""" Helper functions that does not fit in a specific module of the project """
import functools
from typing import Callable, Iterable

def cached_property(method: Callable):
    """Decorator meant to mimic Python 3.8's cached_property"""
    return property(functools.lru_cache()(method))

def find(collection, condition):
    """Return the first element in collection that satisfies condition"""
    return next((element for element in collection if condition(element)), None)

def group_by(values: Iterable, key: Callable, transform: Callable = None):
    """Receives an iterable object and return a dict grouping the elements
    whose the property generated by the 'key' argument are equal as a
    transformed element, if transform is not None"""
    if transform is None:
        transform = lambda item: item
    result: dict = {}
    for value in values:
        _key = key(value)
        result.setdefault(_key, [])
        result[_key].append(transform(value))
    return result



""" Run the application """
from typing import Callable, Dict, List
from collections import namedtuple
from .schema import Schema, Entity
from .validation_entity import needs_validation, canonical_name

Fail = namedtuple('Fail', ['group', 'report_name', 'message'])

def validate_entities(validations: Dict[str, List[Callable]], ignore_tables: List[str], schema: Schema):
    """ Run the schema validation and return a report """
    report: list = []
    for entity in schema.entities():
        if not needs_validation(entity, ignore_tables):
            continue
        group = entity.__name__.lower()
        messages = validate_entity(validations, entity)
        report += [Fail(group, canonical_name(entity), message)
                   for message in messages if message is not None]

    return report

def validate_entity(validations: Dict[str, List[Callable]], entity: Entity):
    """Validate Entity"""
    group = entity.__name__.lower()
    return (_guard_validation(validation, entity)
            for validation in validations.get(group, []))

def _guard_validation(validation: Callable, entity: Entity):
    try:
        return validation(entity)
    except Exception as err: # pylint: disable=broad-except
        return f'validation "{validation.__name__}" raised a {type(err).__name__} with the message "{err}"'



""" A Mock for a Database Adapter, used to run tests """
from typing import List, Dict, Tuple, Set
from .adapter import AbstractAdapter

FILTER_OUT_TABLE_PARAMS = {'triggers', 'columns'}
FILTER_OUT_COLUMN_PARAMS = {'indexes', 'constraints', 'primary_key', 'references'}

class SerializedAdapter(AbstractAdapter):
    """Adapter to a generated serialized schema, that is generated by the validator.
    The Serialized version can be exported to a text file in JSON format and used as caching
    (Not yet implemented)"""
    def __init__(self, info: dict):
        self._info = info

        # caches which methods where called once
        self._called: Set[str] = set()

    def tables(self) -> List[Dict[str, str]]:
        return [
            {'name': table, **{k: v for k, v in params.items() if k not in FILTER_OUT_TABLE_PARAMS}}
            for table, params in self._info.get('tables', {}).items()
        ]

    def _table_cols(self):
        for table, table_params in self._info.get('tables', {}).items():
            for column, column_params in table_params.get('columns', {}).items():
                yield table, column, column_params

    def _reference_columns(self):
        for table, column, column_params in self._table_cols():
            references = column_params.get('references')
            if references:
                yield table, column, references

    def columns(self) -> List[Dict[str, str]]:
        columns_info = []
        for table, column, column_params in self._table_cols():
            params = {k: v for k, v in column_params.items() if k not in FILTER_OUT_COLUMN_PARAMS}
            columns_info.append({'table_name': table, 'name': column, **params})
        return columns_info

    def primary_keys(self) -> List[Tuple[str,str]]:
        return [
            (table, column) for table, column, params in self._table_cols()
            if params.get('primary_key')
        ]

    def references(self) -> List[Tuple[str, str, str]]:
        return list(self._reference_columns())

    def indexes(self) -> List[Dict[str, str]]:
        indexes = []
        for table, column, params in self._table_cols():
            for index, index_params in params.get('indexes', {}).items():
                indexes.append({'table_name': table, 'column_name': column, 'name': index, **index_params})
        return indexes

    def constraints(self) -> List[Dict[str, str]]:
        constraints = []
        for table, column, params in self._table_cols():
            for constraint, cons_params in params.get('constraints', {}).items():
                constraints.append({'table_name': table, 'column_name': column, 'name': constraint, **cons_params})
        return constraints

    def triggers(self) -> List[Dict[str, str]]:
        triggers = []
        for table, params in self._info.get('tables', {}).items():
            for trigger, trigger_params in params.get('triggers', {}).items():
                triggers.append({'table_name': table, 'name': trigger, **trigger_params})
        return triggers

    def _entities(self, group):
        return [{'name': entity, **params} for entity, params in self._info.get(group, {}).items()]

    def functions(self) -> List[Dict[str, str]]:
        return self._entities('functions')

    def procedures(self) -> List[Dict[str, str]]:
        return self._entities('procedures')

    def sequences(self) -> List[Dict[str, str]]:
        return self._entities('sequences')

    # adds a behavior to __getattribute__ that disallows calling a function more than once
    # in order to guarantee that the program will avoid making multiple calls to the adapter
    # when generating the schema
    def __getattribute__(self, attr):
        called = super().__getattribute__('_called')
        if attr in called:
            raise AttributeError(f'Property "{attr}" cannot be called more than once')

        if attr[0] != '_':
            called.add(attr)
        return super().__getattribute__(attr)



"""Setup File"""
from setuptools import setup, find_packages

def long_description(): # pylint: disable=missing-function-docstring
    with open('README.md', 'r') as file:
        return file.read()

setup(
    name='sql_judge',
    version='0.2.0',
    description='Validate SQL database Schemas using Python',
    long_description=long_description(),
    long_description_content_type='text/markdown',
    author='MxBromelia',
    packages= find_packages(where='src'),
    package_dir={"": "src"},
    package_data={
        '': ['*.json']
    },
    entry_points={
        'console_scripts': ['sql-judge=sql_judge.__main__:main']
    },
    python_requires='>=3.7'
)



#! /usr/bin/python3
""" Main Module """
import sys
import logging
import pkg_resources
from .validate import validate_entities
from .parse_configuration.build_configuration import ConfigurationBuilder
from .schema import Schema
from .export import formatted_output

def default_config():
    """default configuration file"""
    return pkg_resources.resource_string(
        'sql_judge.parse_configuration', 'default_configuration.json')

def user_config(filename):
    """user-provided configuration file"""
    try:
        with open(filename) as file:
            return file.read(None)
    except FileNotFoundError as fnf:
        raise RuntimeError(f"File '{filename}' could not be found") from fnf

def sql_judge(filenames) -> list:
    """ Main function """
    if len(filenames) == 0:
        logging.error('At least one configuration file must be provided')
        return []

    try:
        config_builder = ConfigurationBuilder.from_json(default_config())
        for filename in filenames:
            config_builder = config_builder.merge(ConfigurationBuilder.from_json(user_config(filename)))

        config = config_builder.build()
        schema = Schema(config.connection)
        report = validate_entities(config.validations, config.ignore_tables, schema)
        return formatted_output(report, config.export)
    except RuntimeError as err:
        logging.error(str(err))
        return []

def main():
    """Console Scripts entry point"""
    output = sql_judge(sys.argv[1:])

    for line in output:
        print(line)


if __name__ == '__main__':
    main()



"""Validates the Schema"""
from dataclasses import dataclass, field
from typing import Callable, List, Dict
from .adapter import AbstractAdapter

def validates(entity):
    """ Sets Entity to which the validation will run """
    def _validates(validation):
        validation.validates = entity
        return validation
    return _validates

@dataclass
class Configuration:
    """ Stores and configuration options for running the validations """
    connection: AbstractAdapter
    validations: Dict[str, List[Callable]] = field(default_factory=dict)
    ignore_tables: List[str] = field(default_factory=list)
    export: str = 'CLI'



"""
Capture and organize schema validations
"""
import inspect
from collections import defaultdict
VALID_ENTITITES = [
    'table', 'column', 'trigger', 'constraint', 'function', 'procedure', 'index'
]

def _module_functions(module):
    return (element for (_, element) in inspect.getmembers(module, predicate=inspect.isfunction))

def module_validations(module):
    """ Extracts methods decorated as validations from the module """
    return [
        function for function in _module_functions(module)
        if getattr(function, 'validates', None) is not None
    ]

def to_configuration(validations):
    """ Converts a collection of validation functions to the configuration format """
    validations_dict = defaultdict(list)

    for validation in validations:
        validations_dict[validation.validates].append(validation)
    return validations_dict



"""Functions that presents relevant information
about an schema entity regarding validations -
namely: whether it needs validations and the name it will appear on the report"""
from functools import singledispatch
from .schema import Entity, Table, TableEntity, ColumnEntity

@singledispatch
def needs_validation(entity, _ignore_tables: list):
    """Check if entity should or not be validated, considering the ignore-tables configuration"""
    raise TypeError("This method does not accept instances other than Entity")

@needs_validation.register
def _(_entity: Entity, _ignore_tables: list):
    return True

@needs_validation.register # type: ignore
def _(table: Table, ignore_tables: list):
    return table.name not in ignore_tables

@needs_validation.register #type: ignore
def _(table_entity: TableEntity, ignore_tables: list):
    return needs_validation(table_entity.table, ignore_tables)

@singledispatch
def canonical_name(entity):
    """The entity identifier on the report"""
    raise NotImplementedError

@canonical_name.register #type: ignore
def _(entity: Entity):
    return entity.name

@canonical_name.register #type: ignore
def _(table_entity: TableEntity):
    return f'{canonical_name(table_entity.table)}.{table_entity.name}'

@canonical_name.register #type: ignore
def _(column_entity: ColumnEntity):
    return f'{canonical_name(column_entity.column)}.{column_entity.name}'
